@@ -4,6 +4,72 @@ from typing import Any, cast
 import sys
 import threading
 
+# Debug runtime support - will be set by debug_runtime module
+debug_runtime_instance = None
+
+# Track the actual current vars (updated when entering/exiting functions)
+_current_runtime_vars = vars
+
+# Runtime context for error formatting
+_runtime_file: str | None = None
+_runtime_source: str | None = None
+_runtime_current_line: int = 0
+_runtime_current_func: str = '<module>'
+_runtime_current_char: int = 0  # Character position for error reporting
+
+def set_debug_runtime(runtime):
+    """Set the debug runtime instance (called from debug_runtime module)"""
+    global debug_runtime_instance
+    debug_runtime_instance = runtime
+
+def get_debug_runtime():
+    """Get the current debug runtime instance"""
+    return debug_runtime_instance
+
+def get_current_vars():
+    """Get the current variable scope (for debugging)"""
+    return _current_runtime_vars
+
+def get_runtime_context():
+    """Get the current runtime context for error formatting"""
+    return {
+        'file': _runtime_file,
+        'source': _runtime_source,
+        'line': _runtime_current_line,
+        'func': _runtime_current_func,
+        'char': _runtime_current_char
+    }
+
+def set_error_char(char: int):
+    """Set the character position for the next error"""
+    global _runtime_current_char
+    _runtime_current_char = char
+
+def format_runtime_exception(e: Exception) -> str:
+    """Format a runtime exception similar to parse-time exceptions"""
+    ctx = get_runtime_context()
+    error_msg = str(e)
+    char_pos = ctx['char']
+    
+    # Get the source line if available
+    source_line = ""
+    if ctx['source'] and ctx['line'] > 0:
+        lines = ctx['source'].split('\n')
+        if 0 < ctx['line'] <= len(lines):
+            source_line = lines[ctx['line'] - 1]
+    
+    # Format location
+    location = f"{ctx['file']}:{ctx['line']}:{char_pos}" if ctx['file'] else f"Line {ctx['line']}:{char_pos}"
+    
+    # Build error message in same format as parse errors
+    if source_line:
+        pointer = ' ' * char_pos + '^'
+        formatted = f"Runtime Error\n  File \"{ctx['file']}\" line {ctx['line']} in {ctx['func']}\n      {source_line}\n      {pointer}\n    {location}: {error_msg}"
+    else:
+        formatted = f"Runtime Error\n  File \"{ctx['file']}\" line {ctx['line']} in {ctx['func']}\n    {location}: {error_msg}"
+    
+    return formatted
+
 sys.setrecursionlimit(1000000000)
 sys.set_int_max_str_digits(100000000)
 
@@ -18,6 +84,10 @@ builtin_func_names = set(funcs.keys())
 # Lock for thread-safe access to global state during parallel test execution
 _runtime_lock = threading.Lock()
 
+# Track Python imports for alias resolution
+# Maps alias/name -> {'module': str, 'type': 'module'|'name', 'name': str (optional)}
+py_imports: dict[str, dict[str, Any]] = {}
+
 # Constants for magic strings
 NODE_TYPE_STRUCT_DEF = 'struct_def'
 NODE_TYPE_BUILTIN = 'builtin'
@@ -28,14 +98,17 @@ NODE_TYPE_IF = 'if'
 NODE_TYPE_SWITCH = 'switch'
 NODE_TYPE_FOR = 'for'
 NODE_TYPE_FOR_IN = 'for_in'
+NODE_TYPE_TRY = 'try'
 NODE_TYPE_WHILE = 'while'
 NODE_TYPE_BREAK = 'break'
 NODE_TYPE_CONTINUE = 'continue'
 NODE_TYPE_ASSERT = 'assert'
 NODE_TYPE_RETURN = 'return'
+NODE_TYPE_RAISE = 'raise'
 NODE_TYPE_INDEX_ASSIGN = 'index_assign'
 NODE_TYPE_FIELD_ASSIGN = 'field_assign'
 NODE_TYPE_STRING = 'string'
+NODE_TYPE_PY_IMPORT = 'py_import'
 
 # Control flow exceptions for break/continue
 class BreakException(Exception):
@@ -60,9 +133,7 @@ def is_ast_node(obj: Any) -> bool:
 def extract_value(node: Any) -> Any:
     """Extract the actual value from an AST node or return the value as-is"""
     if isinstance(node, dict):
-        if 'value' in node:
-            return node['value']
-        return node
+        return node['value'] if 'value' in node else node
     return node
 
 def ensure_int(value: Any, context: str = "value") -> int:
@@ -72,22 +143,19 @@ def ensure_int(value: Any, context: str = "value") -> int:
     if isinstance(value, (float, str, bool)):
         try:
             return int(value)
-        except (ValueError, TypeError):
-            raise RuntimeError(f"{context} must be numeric, got {type(value).__name__}")
+        except (ValueError, TypeError) as e:
+            raise RuntimeError(
+                f"{context} must be numeric, got {type(value).__name__}"
+            ) from e
     raise RuntimeError(f"{context} must be numeric, not {type(value).__name__}")
 
 def _create_struct_instance(struct_def: dict, args: list) -> dict:
     """Create an instance of a struct from constructor arguments"""
     fields = struct_def['fields']
-    instance = {}
-    
-    for i, field in enumerate(fields):
-        if i < len(args):
-            instance[field['name']] = extract_value(args[i])
-        else:
-            instance[field['name']] = None
-    
-    return instance
+    return {
+        field['name']: extract_value(args[i]) if i < len(args) else None
+        for i, field in enumerate(fields)
+    }
 
 def _process_function_arg(arg: Any, old_vars: dict) -> Any:
     """Process a single function argument, extracting its value"""
@@ -130,14 +198,16 @@ def _bind_function_args(func_args: list, arg_values: list):
             }
 
 def run_func(name: str, args, level=0) -> int|float|bool|str|None|dict|list|Any:
+    # sourcery skip: avoid-builtin-shadow
     from parser import cast_args
-    global vars, runtime
+    global vars, runtime, _current_runtime_vars, _runtime_current_func, _runtime_current_line
     old_vars = vars
+    old_func = _runtime_current_func
+    old_line = _runtime_current_line  # Save calling line
 
     # Check if this is a struct constructor call
     if name in vars and is_struct_def(vars[name]):
         return _create_struct_instance(vars[name], args)
-
     if name not in funcs:
         raise RuntimeError(f'Function "{name}" is not defined.')
 
@@ -151,30 +221,47 @@ def run_func(name: str, args, level=0) -> int|float|bool|str|None|dict|list|Any:
     # Process all arguments to extract their values
     processed_args = [_process_function_arg(arg, old_vars) for arg in args]
 
+    # Only notify debugger for user-defined functions (not built-ins)
+    is_builtin = func['type'] == NODE_TYPE_BUILTIN
+    debug_runtime = get_debug_runtime()
+    if debug_runtime and not is_builtin:
+        debug_runtime.notify_call(name)
+
     # Only clear vars for user-defined functions, preserving struct definitions
-    if func['type'] != NODE_TYPE_BUILTIN:
+    if not is_builtin:
         vars = _preserve_struct_definitions()
+        _current_runtime_vars = vars  # Update current vars for debugger
+        _runtime_current_func = name  # Track function for error reporting
 
     # Bind function arguments to the scope
     if isinstance(func.get('args'), list):
-        _bind_function_args(func['args'], processed_args)
+        _bind_function_args(cast(list, func['args']), processed_args)
 
     # Execute builtin or user-defined function
     if func['type'] == NODE_TYPE_BUILTIN and callable(func['func']):
         out = func['func'](*processed_args)
-        vars = old_vars
-        return out
     else:
         func_body = func['func']
         if not isinstance(func_body, list):
             raise RuntimeError('Invalid function definition.')
-        
+
         if not all(isinstance(item, dict) for item in func_body):
             raise RuntimeError('Invalid function body.')
 
         out = run_scope(cast(AstType, func_body), level+1)
-        vars = old_vars
-        return out
+
+    vars = old_vars
+    if not is_builtin:
+        _current_runtime_vars = vars  # Only restore if we changed it
+        _runtime_current_func = old_func  # Restore function name
+    _runtime_current_line = old_line  # Always restore calling line
+
+    # Notify debugger of function return (only for user-defined functions)
+    debug_runtime = get_debug_runtime()
+    if debug_runtime and not is_builtin:
+        debug_runtime.notify_return()
+
+    return out
 
 class CompiletimeEvalException(Exception): ...
 
@@ -183,17 +270,55 @@ def _get_indexed_value(target: Any, index: Any) -> Any:
     # Ensure index is numeric
     if not isinstance(index, int):
         index = ensure_int(index, "Index")
-    
+
     if isinstance(target, (list, str)):
         try:
             return target[index]
         except (IndexError, TypeError) as e:
-            raise RuntimeError(f"Index error: {e}")
-    
+            # Try to find the index operation in the source for better error reporting
+            ctx = get_runtime_context()
+            if ctx['source'] and ctx['line'] > 0:
+                lines = ctx['source'].split('\n')
+                if 0 < ctx['line'] <= len(lines):
+                    source_line = lines[ctx['line'] - 1]
+                    # Try to find the specific index value in the source
+                    # Look for patterns like [index] where index matches our error value
+                    import re
+                    # Handle negative indices - search for the literal value or the negative form
+                    search_patterns = [f'[{index}]', f'[-{abs(index)}]'] if index < 0 else [f'[{index}]']
+
+                    char_pos = None
+                    for pattern in search_patterns:
+                        pos = source_line.find(pattern)
+                        if pos >= 0:
+                            # Point to the last digit before ']'
+                            char_pos = pos + len(pattern) - 2
+                            break
+
+                    # Fallback: if we can't find the exact index, look for any bracket pair
+                    if char_pos is None:
+                        bracket_pos = source_line.find('[')
+                        if bracket_pos >= 0:
+                            close_bracket = source_line.find(']', bracket_pos)
+                            char_pos = close_bracket - 1 if close_bracket > bracket_pos else bracket_pos
+                    if char_pos is not None:
+                        set_error_char(char_pos)
+            # Create detailed error message
+            error_msg = f"Index error: list index out of range: {index}"
+            if isinstance(target, list):
+                error_msg += f" (length: {len(target)})"
+            raise RuntimeError(error_msg) from e
+
     raise RuntimeError(f"Cannot index into type {type(target).__name__}")
 
 def _handle_list_indexing(node: dict) -> Any:
     """Handle list/string indexing: arr[index]"""
+    global _runtime_current_line
+    
+    # Update current line if this node has line info
+    if 'line' in node:
+        _runtime_current_line = node['line']
+    
     target = eval_expr_node(node['value'])
     index = eval_expr_node(node['slice'])
     
@@ -208,21 +333,38 @@ def _handle_list_indexing(node: dict) -> Any:
     return _get_indexed_value(target, index)
 
 def _handle_field_access(node: dict) -> Any:
-    """Handle struct field access: obj.field"""
+    """Handle struct field access or Python object attribute access: obj.field"""
     target = eval_expr_node(node['value'])
     field_name = node['attr']
-    
+
     # If target is a variable name, look it up
     if isinstance(target, str) and target in vars:
         obj = vars[target].get('value')
+
+        # Check if it's a Python object
+        if hasattr(obj, '__dict__') or hasattr(obj, field_name):
+            # Python object - use getattr
+            try:
+                return getattr(obj, field_name)
+            except AttributeError as e:
+                raise RuntimeError(f"Python object has no attribute '{field_name}'") from e
+
+        # Struct field access
         if isinstance(obj, dict) and field_name in obj:
             return obj[field_name]
         raise RuntimeError(f"Object '{target}' has no field '{field_name}'")
-    
+
+    # If target is a Python object directly, access the attribute
+    if hasattr(target, '__dict__') or hasattr(target, field_name):
+        try:
+            return getattr(target, field_name)
+        except AttributeError as exc:
+            raise RuntimeError(f"Python object has no attribute '{field_name}'") from exc
+
     # If target is already an object (dict), access the field
     if isinstance(target, dict) and field_name in target:
         return target[field_name]
-    
+
     raise RuntimeError(f"Cannot access field '{field_name}' on {target}")
 
 def _handle_fstring(node: dict) -> str:
@@ -247,23 +389,23 @@ def _handle_unary_op(node: dict) -> Any:
     """Handle unary operations (-, +, not)"""
     operand = eval_expr_node(node.get('operand', {}))
     op = node['op']
-    
-    if op == 'USub':
-        if isinstance(operand, (int, float)):
-            return -operand
-        raise RuntimeError(f"Cannot negate {type(operand).__name__}")
+
+    if op == 'Not':
+        return not operand
+
     elif op == 'UAdd':
         if isinstance(operand, (int, float)):
             return +operand
         raise RuntimeError(f"Cannot apply unary + to {type(operand).__name__}")
-    elif op == 'Not':
-        return not operand
-    
+    elif op == 'USub':
+        if isinstance(operand, (int, float)):
+            return -operand
+        raise RuntimeError(f"Cannot negate {type(operand).__name__}")
     return operand
 
 def eval_expr_node(node) -> int|float|bool|str|None|Any:
     from parser import get_type
-    
+
     if not isinstance(node, dict):
         return node
 
@@ -280,7 +422,7 @@ def eval_expr_node(node) -> int|float|bool|str|None|Any:
         return _handle_list_indexing(node)
 
     # F-string (JoinedStr node)
-    if 'values' in node and not ('left' in node or 'op' in node):
+    if 'values' in node and 'left' not in node and 'op' not in node:
         return _handle_fstring(node)
 
     # Variable reference
@@ -289,13 +431,52 @@ def eval_expr_node(node) -> int|float|bool|str|None|Any:
         if value in vars:
             var_value = vars[value].get('value')
             return var_value if var_value is not None else value
+        # At parse time, if variable not found, return the node unchanged
+        # This prevents treating undefined variables as string literals
+        if not runtime:
+            return node
         return value
+    
+    # Function call (from parse_expr - has 'func' key)
+    # OR call from f-string expansion (has 'name' key)
+    if 'func' in node or ('name' in node and 'type' in node and node['type'] == 'call'):
+        # Extract func and args based on node structure
+        if 'func' in node:
+            func = node['func']
+            args = node['args']
+        else:
+            # F-string expansion format: {type: 'call', name: 'str', args: [...]}
+            func = node['name']
+            args = node.get('args', [])
 
-    # Function call
-    if 'func' in node:
-        func = node['func']
-        args = node['args']
-        
+        # Check if this is a method call (obj.method())
+        if isinstance(func, dict) and 'attr' in func and 'value' in func:
+            # This is a method call: obj.method(args)
+            # Get the object
+            obj = eval_expr_node(func['value'])
+            method_name = func['attr']
+
+            # If obj is a variable name, check if it's a module alias first
+            if isinstance(obj, str):
+                if obj in py_imports:
+                    # This is a module function call: ui.Window() where ui is an imported module
+                    # Convert to py_call
+                    from builtin_funcs import funcs as builtin_funcs
+                    py_call_func = builtin_funcs['py_call']['func']
+                    return py_call_func(obj, method_name, *[eval_expr_node(arg) for arg in args]) # type: ignore
+                elif obj in vars:
+                    obj = vars[obj].get('value')
+
+            if not hasattr(obj, '__dict__') and not hasattr(obj, method_name):
+                raise RuntimeError(f"Object has no method '{method_name}'")
+
+            # Python object - call method directly
+            method = getattr(obj, method_name)
+            if not callable(method):
+                raise RuntimeError(f"'{method_name}' is not a callable method")
+            evaluated_args = [eval_expr_node(arg) for arg in args]
+            return method(*evaluated_args)
+        # Regular function call
         # Prepare arguments
         new_args = []
         for arg in args:
@@ -309,9 +490,17 @@ def eval_expr_node(node) -> int|float|bool|str|None|Any:
                 arg = eval_expr_node(arg)
                 arg = {"value": arg, "type": get_type(arg)}
             new_args.append({"value": arg['value'], "type": get_type(arg['value'])})
-        
-        # Get function name
-        func_id = func.get('id') if isinstance(func, dict) else func
+
+        # Get function name - could be 'func' or 'name' key
+        func_id = None
+        if isinstance(func, dict):
+            func_id = func.get('id')
+        elif isinstance(func, str):
+            func_id = func
+        else:
+            # Check if node has 'name' key directly (from f-string expansion)
+            func_id = node.get('name')
+            
         if isinstance(func_id, str):
             return run_func(func_id, new_args)
         raise RuntimeError(f"Invalid function call: {func}")
@@ -320,6 +509,10 @@ def eval_expr_node(node) -> int|float|bool|str|None|Any:
     if 'left' in node:
         return eval_expr(node)
 
+    # Boolean operations (And, Or) with multiple values
+    if 'op' in node and 'values' in node and node['op'] in ('And', 'Or'):
+        return eval_expr(node)
+
     # Field access: obj.field
     if 'attr' in node and 'value' in node:
         return _handle_field_access(node)
@@ -328,6 +521,12 @@ def eval_expr_node(node) -> int|float|bool|str|None|Any:
     if 'type' in node and node['type'] == NODE_TYPE_STRING:
         return node.get('value', '')
 
+    # Set literal: {'type': 'set', 'value': [...]}
+    if 'type' in node and node['type'] == 'set':
+        # Convert to Python set, evaluating each element
+        elements = node.get('value', [])
+        return set(eval_expr_node(elem) for elem in elements)
+
     # Direct value
     if 'value' in node:
         return node['value']
@@ -340,7 +539,7 @@ def eval_expr_node(node) -> int|float|bool|str|None|Any:
     if isinstance(node, dict):
         # Check if it looks like a struct instance (no AST keys like 'func', 'op', 'args', etc.)
         ast_keys = {'func', 'op', 'ops', 'left', 'right', 'args', 'type', 'id', 'attr', 'slice', 'value', 'values'}
-        if not any(key in node for key in ast_keys):
+        if all(key not in node for key in ast_keys):
             # Looks like a struct instance - return as-is
             return node
         if 'mods' in node:
@@ -364,10 +563,14 @@ def eval_expr_calc(left, op, right):
         return left >= right
     elif op == 'LtE':
         return left <= right
+    elif op == 'In':
+        return left in right
+    elif op == 'NotIn':
+        return left not in right
     elif op == 'Or':
-        return left or right
+        return bool(left or right)
     elif op == 'And':
-        return left and right
+        return bool(left and right)
     elif op == 'Not':
         return not left
     elif op == 'BitXor':
@@ -377,18 +580,17 @@ def eval_expr_calc(left, op, right):
     elif op == 'BitOr':
         return left | right
 
-    # Math
-    elif op == 'Add' or op == '+':
+    elif op in ['Add', '+']:
         return left + right
-    elif op == 'Sub' or op == '-':
+    elif op in ['Sub', '-']:
         return left - right
-    elif op == 'Mult' or op == '*':
+    elif op in ['Mult', '*']:
         return left * right
-    elif op == 'Div' or op == '/':
+    elif op in ['Div', '/']:
         return left / right
-    elif op == 'Pow' or op == '**':
+    elif op in ['Pow', '**']:
         return left** right
-    elif op == 'Mod' or op == '%':
+    elif op in ['Mod', '%']:
         return left % right
     else:
         raise RuntimeError(f'Invalid operator: {op}')
@@ -419,15 +621,12 @@ def eval_expr(expr, level=0) -> int|float|bool|str|None:
         ops = [expr['op']]
         if 'operand' in expr:
             return _handle_unary_op(expr)
-        else:
-            left = expr['values'][0]
-            comps = expr['values'][1:]
-    # Handle comparison chains
+        left = eval_expr_node(expr['values'][0])
+        comps = expr['values'][1:]
     elif 'ops' in expr:
         left = eval_expr_node(expr['left'])
         ops = expr['ops']
         comps = expr['comparators']
-    # Handle binary operations
     elif 'op' in expr:
         left = eval_expr_node(expr['left'])
         ops = [expr['op']]
@@ -438,7 +637,7 @@ def eval_expr(expr, level=0) -> int|float|bool|str|None:
     # Evaluate comparison/operation chain
     for op, comp in zip(ops, comps):
         comp = eval_expr_node(comp)
-        
+
         # Resolve variable references
         if isinstance(left, str) and left in vars:
             left_old = left
@@ -451,9 +650,11 @@ def eval_expr(expr, level=0) -> int|float|bool|str|None:
 
         try:
             left = eval_expr_calc(left, op, comp)
-        except:
+        except Exception as e:
             if isinstance(left, str):
-                raise SyntaxError(f'{left} is not defined.')
+                raise SyntaxError(f'{left} is not defined.') from e
+            # Re-raise all other exceptions (ZeroDivisionError, etc.)
+            raise
 
     return left
 
@@ -487,20 +688,58 @@ def _execute_node_struct_def(node: dict):
         "fields": node['fields']
     }
 
+def _execute_node_py_import(node: dict):
+    """Handle py_import node"""
+    module_name = node.get('module', '')
+    alias = node.get('alias')
+    name = node.get('name')
+    
+    if not module_name:
+        return
+    
+    # Call the py_import builtin to import the module
+    run_func('py_import', [module_name])
+    
+    # Track the import for alias resolution
+    if name:
+        # "from module py_import name" or "from module py_import name as alias"
+        key = alias or name
+        py_imports[key] = {
+            'module': module_name,
+            'type': 'name',
+            'name': name
+        }
+    else:
+        # "py_import module" or "py_import module as alias"
+        key = alias or module_name
+        py_imports[key] = {
+            'module': module_name,
+            'type': 'module'
+        }
+
 def _execute_node_var(node: dict):
     """Handle variable declaration node"""
     from parser import cast_value
+    import copy
     
     value = cast_value(node['value'], node['value_type'])
     
     # Evaluate f-strings and expressions
     if isinstance(value, dict):
-        if 'values' in value and 'value' not in value:
+        # Check if it's a set/list literal that needs evaluation
+        if value.get('type') in ('set', 'list'):
+            value = eval_expr_node(value)
+        # Check if it's a runtime expression (has slice, attr, or other runtime features)
+        elif 'slice' in value or 'attr' in value or ('values' in value and 'value' not in value):
             value = eval_expr_node(value)
         elif 'value' not in value:
             value = eval_expr_node(value)
         else:
             value = value['value']
+    
+    # Make a deep copy of mutable objects to avoid mutating the AST
+    if isinstance(value, (list, dict, set)):
+        value = copy.deepcopy(value)
 
     vars[node['name']] = {
         "type": node['value_type'],
@@ -527,13 +766,39 @@ def _execute_node_index_assign(node: dict):
 def _execute_node_field_assign(node: dict):
     """Handle struct field assignment: obj.field = value"""
     target_name = node['target']
+    
+    # Check if it's a Python module import alias
+    if target_name in py_imports:
+        # Setting attribute on a Python module
+        field_name = node['field']
+        value = eval_expr(node['value'])
+        # Use setattr to set the module attribute
+        import sys
+        import_info = py_imports[target_name]
+        module_name = import_info['module']
+        
+        # Import the module to get reference
+        if module_name not in sys.modules:
+            __import__(module_name)
+        module = sys.modules[module_name]
+        setattr(module, field_name, value)
+        return
+    
     if target_name not in vars:
         raise RuntimeError(f"Variable '{target_name}' is not defined")
     
     target = vars[target_name]['value']
+    target_type = vars[target_name]['type']
     field_name = node['field']
     value = eval_expr(node['value'])
     
+    # Handle pyobject types using py_setattr
+    if target_type in ('pyobject', 'pyobj'):
+        from builtin_funcs import funcs as builtin_funcs
+        py_setattr_func = cast(Any, builtin_funcs['py_setattr']['func'])
+        py_setattr_func(target, field_name, value)
+        return
+    
     if isinstance(target, dict) and field_name in target:
         target[field_name] = value
     else:
@@ -541,25 +806,26 @@ def _execute_node_field_assign(node: dict):
 
 def _execute_node_if(node: dict) -> Any:
     """Handle if/elif/else statement"""
-    if eval_expr(node['condition']):
-        return run_scope(node['scope'])
+    condition_result = eval_expr(node['condition'])
     
+    if condition_result:
+        return run_scope(node['scope'])
+
     # Try elif branches
     if node['elifs']:
         for elif_node in node['elifs']:
             if eval_expr(elif_node['condition']):
                 return run_scope(elif_node['scope'])
-    
+
     # Execute else if present
     if node['else']:
         return run_scope(node['else'])
-    
     return None
 
 def _execute_node_switch(node: dict) -> Any:
     """Handle switch statement"""
     switch_value = eval_expr(node['expr'])
-    
+
     # Try to match against each case
     for case in node['cases']:
         for case_val_node in case['values']:
@@ -567,12 +833,9 @@ def _execute_node_switch(node: dict) -> Any:
             if switch_value == case_val:
                 # Execute the case body and return immediately
                 return run_scope(case['body'])
-    
+
     # Execute default if no match
-    if node['default']:
-        return run_scope(node['default'])
-    
-    return None
+    return run_scope(node['default']) if node['default'] else None
 
 def _execute_node_for(node: dict, level: int):
     """Handle for loop with range"""
@@ -635,18 +898,92 @@ def _execute_node_assert(node: dict):
             raise AssertionError(message)
         raise AssertionError('Assertion failed')
 
+def _execute_node_try(node: dict, level: int) -> Any:
+    """Handle try-except statement"""
+    try:
+        # Execute the try block
+        result = run_scope(node['try_scope'], level+1)
+        return result
+    except Exception as e:
+        # Get the exception type name
+        exc_name = type(e).__name__
+        expected_exc = node['exc_type']
+        
+        # Check if the exception type matches
+        if exc_name == expected_exc:
+            # Execute the except block
+            result = run_scope(node['except_scope'], level+1)
+            return result
+        else:
+            # Re-raise if it doesn't match
+            raise
+
+def _execute_node_raise(node: dict):
+    """Handle raise statement"""
+    global _runtime_current_char
+    
+    exc_type = node.get('exc_type')
+    message = node.get('message', '')
+    
+    # Update character position for error reporting
+    if 'char' in node:
+        _runtime_current_char = node['char']
+    
+    if not exc_type:
+        # Bare raise - re-raise current exception
+        raise
+    
+    # Map exception type string to Python exception class
+    exc_classes = {
+        'ZeroDivisionError': ZeroDivisionError,
+        'ValueError': ValueError,
+        'TypeError': TypeError,
+        'IndexError': IndexError,
+        'KeyError': KeyError,
+        'AttributeError': AttributeError,
+        'RuntimeError': RuntimeError,
+        'AssertionError': AssertionError,
+        'Exception': Exception,
+    }
+    
+    exc_class = exc_classes.get(exc_type, RuntimeError)
+    
+    # Include exception type in message for better error reporting
+    full_message = f"[{exc_type}] {message}" if message else f"[{exc_type}]"
+    
+    raise exc_class(full_message)
+
 def run_scope(ast: AstType, level=0):
     """Execute a scope (list of AST nodes)"""
+    global _runtime_current_line
     result = None
-    
-    for node in ast:
+
+    for idx, node in enumerate(ast):
+        # Update current line for error context
+        if isinstance(node, dict):
+            _runtime_current_line = node.get('line', idx + 1)
+        else:
+            _runtime_current_line = idx + 1
+            
+        if debug_runtime := get_debug_runtime():
+            debug_runtime.notify_line(_runtime_current_line)
+
+        # Handle call expressions from AST first (e.g. method calls like obj.method())
+        # These don't have a 'type' key, just 'func' and 'args'
+        if 'func' in node and 'args' in node and 'type' not in node:
+            # Execute the expression but discard the result since this is used as a statement
+            result = eval_expr(node)
+            continue
+
         node_type = node['type']
-        
+
         # Handle each node type
         if node_type == NODE_TYPE_FUNCTION:
             _execute_node_function(node, level)
         elif node_type == NODE_TYPE_STRUCT_DEF:
             _execute_node_struct_def(node)
+        elif node_type == NODE_TYPE_PY_IMPORT:
+            _execute_node_py_import(node)
         elif node_type == NODE_TYPE_VAR:
             _execute_node_var(node)
         elif node_type == NODE_TYPE_INDEX_ASSIGN:
@@ -671,6 +1008,10 @@ def run_scope(ast: AstType, level=0):
             _execute_node_for_in(node, level)
         elif node_type == NODE_TYPE_WHILE:
             _execute_node_while(node)
+        elif node_type == NODE_TYPE_TRY:
+            result = _execute_node_try(node, level)
+            if result is not None:
+                return result
         elif node_type == NODE_TYPE_BREAK:
             raise BreakException(node.get('level', 1))
         elif node_type == NODE_TYPE_CONTINUE:
@@ -679,19 +1020,30 @@ def run_scope(ast: AstType, level=0):
             _execute_node_assert(node)
         elif node_type == NODE_TYPE_RETURN:
             return eval_expr(node['value'])
+        elif node_type == NODE_TYPE_RAISE:
+            _execute_node_raise(node)
         else:
             raise RuntimeError(f'Unknown node type: {node_type}')
-    
+
     return result
 
-def run(ast:AstType):
+def run(ast:AstType, file:str='', source:str=''):
     # Use lock to prevent race conditions in parallel test execution
     with _runtime_lock:
-        global vars, runtime, funcs
+        global vars, runtime, funcs, py_imports, _current_runtime_vars
+        global _runtime_file, _runtime_source, _runtime_current_func
+        
+        # Initialize runtime context for error formatting
+        _runtime_file = file
+        _runtime_source = source
+        _runtime_current_func = '<module>'
         
         runtime = True
         # Reset vars to clear any state from previous runs (critical for parallel test execution)
         vars.clear()
+        _current_runtime_vars = vars  # Initialize current vars tracking
+        # Reset py_imports to clear any import aliases from previous runs
+        py_imports.clear()
         # Reset funcs to only include builtin functions (remove user-defined functions from previous runs)
         # Remove any functions that are not in the original builtin set
         user_func_names = [name for name in funcs.keys() if name not in builtin_func_names]
